;; Pentagonal numbers are generated by the formula, Pn=n(3n-1)/2. The
;; first ten pentagonal numbers are:

;; 1, 5, 12, 22, 35, 51, 70, 92, 117, 145, ...

;; It can be seen that P4 + P7 = 22 + 70 = 92 = P8. However, their
;; difference, 70 22 = 48, is not pentagonal.

;; Find the pair of pentagonal numbers, Pj and Pk, for which their sum
;; and difference is pentagonal and D = |Pk Pj| is minimised; what is
;; the value of D?


(defun pentagonal (n)
  (/
   (*
    n
    (- (* 3 n) 1))
   2))

(defun get-pentagonals (count &optional (start 1))
  (loop for ii from start to (+ count start -1)
	collect (pentagonal ii)))

;; (get-pentagonals 10 2)

;; Can determine an upper bound? and can we limit the search in some way?
;; 2*Pn=3n^2-n

(defun memoize (function)
  (let ((res-table (make-hash-table :test #'equal :size 100000)))
    (lambda (&rest args)
      (multiple-value-bind 
	    (value exists) 
	  (gethash args res-table)
	(when (not exists)
	  (setf value (apply function args))
	  (setf (gethash args res-table) value))
	value))))

(defun inner-is-pentagonal (num)
  (loop for ii = 1 then (+ ii 1)
	for pent = (pentagonal ii)
	do
	(when (> pent num)
	  (return nil))
	(when (= pent num)
	  (return t))))

(defvar *inner-pent-mem* (memoize #'inner-is-pentagonal))

(defun is-pentagonal (num)
  (funcall *inner-pent-mem* num))

;; (is-pentagonal 12)

'(loop for ii from 2 to 1000
  for pent = (pentagonal ii) then (pentagonal ii)
  for prev-pent = (pentagonal (- ii 1)) then (pentagonal (- ii 1))
  do
  (let ((sum (+ prev-pent pent))
	(diff (- pent prev-pent)))
    (when (and (is-pentagonal sum)
	       (is-pentagonal diff))
      (format t "ii:~a pent:~a prev-pent:~a sum:~a diff:~a~&"
	      ii pent prev-pent sum diff))))

;; (get-pentagonals 10 1000)


(time
 (loop for pent-list = (get-pentagonals 10000) then (cdr pent-list)
       for pent = (car pent-list)
       while pent
       do
       (format t ".")
       (loop for pent2 in (cdr pent-list)
	     do
	     (let ((sum (+ pent pent2))
		   (diff (- pent2 pent)))
	       (when (and (is-pentagonal sum)
			  (is-pentagonal diff))
		 (format t "~a vs ~a sum:~a diff:~a~&"
			 pent
			 pent2
			 sum
			 diff))))))
